@@ -1,19 +1,3 @@
-# locale ---------------------------------------------------------------------
-
-set_locale <- function(cats) {
-  cats <- as_character(cats)
-  stopifnot(is.named(cats))
-
-  if ("LC_ALL" %in% names(cats)) {
-    stop("Setting LC_ALL category not implemented.", call. = FALSE)
-  }
-
-  old <- vapply(names(cats), Sys.getlocale, character(1))
-
-  mapply(Sys.setlocale, names(cats), cats)
-  invisible(old)
-}
-
 #' Locale settings
 #'
 #' Temporarily change locale settings.
@@ -51,8 +35,50 @@ set_locale <- function(cats) {
 #' with_locale(c(LC_COLLATE = "C"), sort(x))
 #'
 #' @export
-with_locale <- with_(set_locale)
+with_locale <- function(new, code) {
+  local_locale(new)
+  code
+}
 
 #' @rdname with_locale
 #' @export
-local_locale <- local_(set_locale, dots = TRUE)
+local_locale <- function(.new = list(),
+                         ...,
+                         .local_envir = parent.frame()) {
+  new <- list_combine(as.list(.new), list(...))
+  cats <- as_locale_cats(new)
+
+  # <https://github.com/r-lib/withr/issues/179>
+  # R supports setting LC_COLLATE to C via envvar. When that is the
+  # case, it takes precedence over the currently set locale. We need
+  # to set both the envvar and the locale for collate to fully take
+  # effect.
+  if ("LC_COLLATE" %in% names(cats)) {
+    collate <- cats["LC_COLLATE"]
+    local_envvar(collate, .local_envir = .local_envir)
+  }
+
+  old <- get_locale(cats)
+  defer(set_locale(old), envir = .local_envir)
+  set_locale(cats)
+
+  invisible(old)
+}
+
+set_locale <- function(cats) {
+  mapply(Sys.setlocale, names(cats), cats)
+}
+get_locale <- function(cats) {
+  vapply(names(cats), Sys.getlocale, character(1))
+}
+
+as_locale_cats <- function(cats) {
+  cats <- as_character(cats)
+  stopifnot(is.named(cats))
+
+  if ("LC_ALL" %in% names(cats)) {
+    stop("Setting LC_ALL category not implemented.", call. = FALSE)
+  }
+
+  cats
+}
